"""Bits and pieces."""

import inspect
import os
import random
import socket
import time
from dataclasses import dataclass
from typing import Annotated, Callable, Generic, NamedTuple, TypeVar

import typer

DEBUG = os.getenv("DEBUG") not in ["0", "false", "False", None]

T = TypeVar("T")


class Retry:
    """Retry variant."""


class Ok(NamedTuple, Generic[T]):
    """Success variant."""

    result: T


type Result[T] = Ok[T] | RuntimeError | Retry


def log(msg: str, stack_offset: int = 1) -> None:
    fn = inspect.stack()[stack_offset].function
    print(f"[{fn}]: {msg}")


def debug(msg: str) -> None:
    if DEBUG:
        log(msg, 2)


def retries[
    **P
](num_retries: int, sleep_time: float) -> Callable[
    [Callable[P, Result | Retry]], Callable[P, Result]
]:
    """ "just put a retry loop around it" """

    def wrapper(fn: Callable[P, Result | Retry]) -> Callable[P, Result]:
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> Result:
            for _ in range(0, num_retries):
                match fn(*args, **kwargs):
                    case Ok(x):
                        return x
                    case Retry():
                        time.sleep(sleep_time)
                        continue
                    case RuntimeError() as e:
                        return e
            return RuntimeError(f"retries exceeded in '{fn.__name__}'")

        return wrapped

    return wrapper


@dataclass(frozen=True)
class Opts:
    """Holds common options."""

    commit_sha: str
    workflow: str
    gh_user: str
    gh_pass: str
    gh_otp: str
    gh_pat: str
    headless: bool
    port: int | None


def cli(main: Callable[[Opts], None]) -> Callable:
    # pylint: disable=R0913,R0917
    def _inner(
        commit_sha: Annotated[str, typer.Argument(callback=_sha_callback)],
        workflow: str,
        gh_user: Annotated[str, typer.Option(envvar="_GH_USER")],
        gh_pass: Annotated[str, typer.Option(envvar="_GH_PASS")],
        gh_otp: Annotated[str, typer.Option(envvar="_GH_OTP")],
        gh_pat: Annotated[str, typer.Option(envvar="_GH_PAT")],
        headless: Annotated[bool, typer.Option(envvar="_HEADLESS")] = True,
        port: Annotated[int | None, typer.Option(envvar="_PORT")] = None,
    ) -> None:
        opts = Opts(commit_sha, workflow, gh_user, gh_pass, gh_otp, gh_pat, headless, port)
        main(opts)

    return _inner


def find_free_port(min_port: int = 8100, max_port: int = 8500) -> int | None:
    num_tries = 0
    random_port = random.randint(min_port, max_port)
    while not is_port_open(random_port):
        if num_tries > 100:
            return None
        random_port = random.randint(min_port, max_port)
        num_tries += 1
    return random_port


def is_port_open(port: int) -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    res = sock.connect_ex(("127.0.0.1", port))
    sock.close()
    return res != 0


def _sha_callback(value: str) -> str:
    if len(value) != 40:
        raise typer.BadParameter("need a full 40 character long commit sha")
    return value



#!/usr/bin/env -S /bin/sh -c 'exec "$(dirname $(readlink -f "$0"))/../.venv/bin/python3" "$0" "$@"'

"""Having your cake and eating it too."""
import dataclasses
import multiprocessing as mp
import sys
from multiprocessing.synchronize import Lock as LockBase
from pathlib import Path
from threading import Event
from typing import Dict, Union

import typer
from github import Auth, Github
from github.WorkflowJob import WorkflowJob
from github.WorkflowRun import WorkflowRun
from pykka import ActorRegistry, ThreadingActor
from xdg.BaseDirectory import xdg_cache_home

from octotail.browser import BrowseRequest, CloseRequest, ExitRequest, VisitRequest, run_browser
from octotail.gh import JobDone, RunWatcher, WorkflowDone, get_active_run, guess_repo
from octotail.mitm import ProxyWatcher, WsSub
from octotail.streamer import run_streamer
from octotail.utils import Opts, cli, debug, find_free_port, log

COOKIE_JAR = Path(xdg_cache_home) / "octotail" / "gh-cookies.json"

type MgrMessage = Union[WorkflowJob, WsSub, JobDone, WorkflowDone]


class Manager(ThreadingActor):
    """I'm the Baahwss."""

    browse_queue: mp.Queue
    stop: Event
    background_tasks: Dict[int, mp.Process]
    output_lock: LockBase
    job_map: Dict[int, str]

    def __init__(self, browse_queue: mp.Queue, stop: Event):
        super().__init__()
        self.browse_queue = browse_queue
        self.stop = stop
        self.background_tasks = {}
        self.output_lock = mp.Lock()
        self.job_map = {}

    def on_receive(self, msg: MgrMessage) -> None:
        debug(f"manager got message: {msg!r}")

        match msg:
            case WorkflowJob() as job:
                self.browse_queue.put_nowait(VisitRequest(job.html_url, job.id))
                self.job_map[job.id] = job.name

            case WsSub() as ws_sub:
                if ws_sub.job_id in self.job_map:
                    ws_sub = dataclasses.replace(ws_sub, job_name=self.job_map[ws_sub.job_id])
                self.browse_queue.put_nowait(CloseRequest(ws_sub.job_id))
                self.background_tasks[ws_sub.job_id] = run_streamer(ws_sub, self.output_lock)

            case JobDone() as job_done:
                print(f"[{job_done.job_name}]: conclusion: {job_done.conclusion}")
                if (streamer := self.background_tasks.get(job_done.job_id)) is not None:
                    streamer.terminate()
                    del self.background_tasks[job_done.job_id]

            case WorkflowDone() as wf_done:
                print(f"[workflow]: conclusion: {wf_done.conclusion}")
                self.browse_queue.put_nowait(ExitRequest())
                self.stop.set()

    def on_stop(self) -> None:
        log("manager stopping")
        self.browse_queue.put_nowait(ExitRequest())
        for p in self.background_tasks.values():
            p.terminate()


@cli
def main(opts: Opts) -> None:
    _stop = Event()

    if (repo_id := guess_repo()) is None:
        sys.exit(1)

    gh_client = Github(auth=Auth.Token(opts.gh_pat))
    wf_run = get_active_run(gh_client.get_repo(repo_id), opts.commit_sha, opts.workflow)
    if not isinstance(wf_run, WorkflowRun):
        log(f"could not get a matching workflow run: {wf_run}")
        sys.exit(1)

    # pylint: disable=E1136
    browser_inbox: mp.Queue[BrowseRequest] = mp.Queue()

    # find a free port
    if opts.port is None:
        if (port := find_free_port()) is not None:
            opts = dataclasses.replace(opts, port=port)
        else:
            log("giving up finding a free port in the 8100 - 8500 range")
            sys.exit(1)

    debug(f"starting on port {opts.port}")

    mp.Process(target=run_browser, args=(opts, browser_inbox)).start()
    manager = Manager.start(browser_inbox, _stop)
    run_watcher = RunWatcher.start(wf_run, manager, _stop)
    proxy_watcher = ProxyWatcher.start(manager, _stop, opts.port)

    run_watcher_future = run_watcher.proxy().watch()
    proxy_watcher_future = proxy_watcher.proxy().watch()
    try:
        run_watcher_future.join(proxy_watcher_future).get()
    except KeyboardInterrupt:
        _stop.set()

    ActorRegistry.stop_all()


def _main() -> None:
    typer.run(main)


if __name__ == "__main__":
    _main()



from codecs import open
from os import path

from setuptools import find_packages, setup

dot = path.abspath(path.dirname(__file__))

# get the dependencies and installs
with open(path.join(dot, "requirements.txt"), encoding="utf-8") as f:
    all_reqs = f.read().split("\n")

install_requires = [x.strip() for x in all_reqs if "git+" not in x]
dependency_links = [x.strip().replace("git+", "") for x in all_reqs if x.startswith("git+")]

setup(
    name="octotail",
    version="1.0.0",
    description="Live tail GitHub Actions runs on git push.",
    entry_points={
        "console_scripts": ["octotail=octotail.main:_main"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
    ],
    license="UNLICENSE",
    keywords="github-actions, tail, post-receive, codecrafters, git",
    packages=find_packages(exclude=["docs", "tests*"]),
    include_package_data=True,
    install_requires=install_requires,
    dependency_links=dependency_links,
)



"""Streamer actor."""

import asyncio as aio
import json
import multiprocessing as mp
from contextlib import suppress
from multiprocessing.synchronize import Lock as LockBase

import websockets.client

from octotail.browser import RANDOM_UA
from octotail.mitm import WsSub
from octotail.utils import log

WS_HEADERS = {
    "User-Agent": RANDOM_UA,
    "Origin": "https://github.com",
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "websocket",
    "Sec-Fetch-Site": "same-site",
    "Pragma": "no-cache",
    "Cache-Control": "no-cache",
}


def run_streamer(ws_sub: WsSub, lock: LockBase) -> mp.Process:
    p = mp.Process(target=_streamer, args=(ws_sub, lock))
    p.start()
    return p


def _streamer(ws_sub: WsSub, lock: LockBase) -> None:
    loop = aio.new_event_loop()
    aio.set_event_loop(loop)
    try:
        loop.run_until_complete(_stream_it(ws_sub, lock))
    except KeyboardInterrupt:
        loop.close()


async def _stream_it(ws_sub: WsSub, lock: LockBase) -> None:
    ws_url = "wss://" + ws_sub.url.removeprefix("https://")
    job_name = ws_sub.job_name or "unknown"

    websocket = None
    try:
        async with websockets.client.connect(ws_url, extra_headers=WS_HEADERS) as websocket:
            await websocket.send(ws_sub.subs)
            async for msg in websocket:
                lines = _extract_lines(job_name, msg)
                if lines:
                    lock.acquire()
                    print(lines)
                    lock.release()
    except aio.CancelledError:
        log("cancelled")
        if websocket:
            await websocket.close()


def _extract_lines(job_name: str, x: str) -> str:
    with suppress(Exception):
        return "\n".join(
            f"[{job_name}]: {l['line']}" for l in json.loads(x)["data"]["data"]["lines"]
        )
    return ""
